/*
 * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#include "config.h"

#include <fwupdplugin.h>

#include "fu-nvme-device.h"

static void
fu_plugin_nvme_init(FuPlugin *plugin)
{
	fu_plugin_add_udev_subsystem(plugin, "nvme");
	fu_plugin_add_device_gtype(plugin, FU_TYPE_NVME_DEVICE);
}

void
fu_plugin_init_vfuncs(FuPluginVfuncs *vfuncs)
{
	vfuncs->build_hash = FU_BUILD_HASH;
	vfuncs->init = fu_plugin_nvme_init;
}
